"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var handlebars = require("handlebars");
var Type_1 = require("../Type");
var TypeUtils_1 = require("../TypeUtils");
var Source_1 = require("../Source");
var Strings_1 = require("../Strings");
var Support_1 = require("../Support");
var Naming_1 = require("../Naming");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var Annotation_1 = require("../Annotation");
var unicode = require("unicode-properties");
var AccessModifier;
(function (AccessModifier) {
    AccessModifier[AccessModifier["None"] = 0] = "None";
    AccessModifier[AccessModifier["Public"] = 1] = "Public";
    AccessModifier[AccessModifier["Internal"] = 2] = "Internal";
})(AccessModifier = exports.AccessModifier || (exports.AccessModifier = {}));
var CSharpTargetLanguage = /** @class */ (function (_super) {
    __extends(CSharpTargetLanguage, _super);
    function CSharpTargetLanguage() {
        var _this = _super.call(this, "C#", ["cs", "csharp"], "cs") || this;
        _this._listOption = new RendererOptions_1.EnumOption("array-type", "Use T[] or List<T>", [
            ["array", false],
            ["list", true]
        ]);
        _this._denseOption = new RendererOptions_1.EnumOption("density", "Property density", [["normal", false], ["dense", true]], "normal", "secondary");
        _this._featuresOption = new RendererOptions_1.EnumOption("features", "Output features", [
            ["complete", { helpers: true, attributes: true }],
            ["attributes-only", { helpers: false, attributes: true }],
            ["just-types", { helpers: false, attributes: false }]
        ]);
        // FIXME: Do this via a configurable named eventually.
        _this._namespaceOption = new RendererOptions_1.StringOption("namespace", "Generated namespace", "NAME", "QuickType");
        _this._versionOption = new RendererOptions_1.EnumOption("csharp-version", "C# version", [["5", 5], ["6", 6]], "6", "secondary");
        _this._checkRequiredOption = new RendererOptions_1.BooleanOption("check-required", "Fail if required properties are missing", false);
        _this._useDecimalOption = new RendererOptions_1.EnumOption("number-type", "Type to use for numbers", [["double", false], ["decimal", true]], "double", "secondary");
        return _this;
    }
    CSharpTargetLanguage.prototype.getOptions = function () {
        return [
            this._namespaceOption,
            this._versionOption,
            this._denseOption,
            this._listOption,
            this._useDecimalOption,
            this._featuresOption,
            this._checkRequiredOption
        ];
    };
    Object.defineProperty(CSharpTargetLanguage.prototype, "partialStringTypeMapping", {
        get: function () {
            return { date: "date-time", time: "date-time", dateTime: "date-time" };
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CSharpTargetLanguage.prototype, "supportsUnionsWithBothNumberTypes", {
        get: function () {
            return true;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CSharpTargetLanguage.prototype, "supportsOptionalClassProperties", {
        get: function () {
            return true;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CSharpTargetLanguage.prototype, "rendererClass", {
        get: function () {
            return NewtonsoftCSharpRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return CSharpTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
exports.default = CSharpTargetLanguage;
var namingFunction = Naming_1.funPrefixNamer("namer", csNameStyle);
// FIXME: Make a Named?
var denseJsonPropertyName = "J";
var denseRequiredEnumName = "R";
var denseNullValueHandlingEnumName = "N";
function isStartCharacter(utf16Unit) {
    if (unicode.isAlphabetic(utf16Unit)) {
        return true;
    }
    return utf16Unit === 0x5f; // underscore
}
function isPartCharacter(utf16Unit) {
    var category = unicode.getCategory(utf16Unit);
    if (["Nd", "Pc", "Mn", "Mc"].indexOf(category) >= 0) {
        return true;
    }
    return isStartCharacter(utf16Unit);
}
var legalizeName = Strings_1.utf16LegalizeCharacters(isPartCharacter);
function csNameStyle(original) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, "", isStartCharacter);
}
function isValueType(t) {
    if (t instanceof Type_1.UnionType) {
        return TypeUtils_1.nullableFromUnion(t) === null;
    }
    return ["integer", "double", "bool", "enum", "date-time"].indexOf(t.kind) >= 0;
}
var CSharpRenderer = /** @class */ (function (_super) {
    __extends(CSharpRenderer, _super);
    function CSharpRenderer(targetLanguage, graph, leadingComments, namespaceName, _version, dense, _useList, _useDecimal) {
        var _this = _super.call(this, targetLanguage, graph, leadingComments) || this;
        _this.namespaceName = namespaceName;
        _this._version = _version;
        _this.dense = dense;
        _this._useList = _useList;
        _this._useDecimal = _useDecimal;
        return _this;
    }
    CSharpRenderer.prototype.forbiddenNamesForGlobalNamespace = function () {
        return ["QuickType", "Type", "System", "Console", "Exception", "DateTimeOffset"];
    };
    CSharpRenderer.prototype.forbiddenForObjectProperties = function (_, classNamed) {
        return {
            names: [
                classNamed,
                "ToString",
                "GetHashCode",
                "Finalize",
                "Equals",
                "GetType",
                "MemberwiseClone",
                "ReferenceEquals"
            ],
            includeGlobalForbidden: false
        };
    };
    CSharpRenderer.prototype.forbiddenForUnionMembers = function (_, unionNamed) {
        return { names: [unionNamed], includeGlobalForbidden: true };
    };
    CSharpRenderer.prototype.makeNamedTypeNamer = function () {
        return namingFunction;
    };
    CSharpRenderer.prototype.namerForObjectProperty = function () {
        return namingFunction;
    };
    CSharpRenderer.prototype.makeUnionMemberNamer = function () {
        return namingFunction;
    };
    CSharpRenderer.prototype.makeEnumCaseNamer = function () {
        return namingFunction;
    };
    CSharpRenderer.prototype.unionNeedsName = function (u) {
        return TypeUtils_1.nullableFromUnion(u) === null;
    };
    CSharpRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        // If the top-level type doesn't contain any classes or unions
        // we have to define a class just for the `FromJson` method, in
        // emitFromJsonForTopLevel.
        return TypeUtils_1.directlyReachableSingleNamedType(type);
    };
    CSharpRenderer.prototype.emitBlock = function (f, semicolon) {
        if (semicolon === void 0) { semicolon = false; }
        this.emitLine("{");
        this.indent(f);
        this.emitLine("}", semicolon ? ";" : "");
    };
    CSharpRenderer.prototype.csType = function (t, withIssues) {
        var _this = this;
        if (withIssues === void 0) { withIssues = false; }
        return TypeUtils_1.matchType(t, function (_anyType) { return Source_1.maybeAnnotated(withIssues, Annotation_1.anyTypeIssueAnnotation, "object"); }, function (_nullType) { return Source_1.maybeAnnotated(withIssues, Annotation_1.nullTypeIssueAnnotation, "object"); }, function (_boolType) { return "bool"; }, function (_integerType) { return "long"; }, function (_doubleType) { return (_this._useDecimal ? "decimal" : "double"); }, function (_stringType) { return "string"; }, function (arrayType) {
            var itemsType = _this.csType(arrayType.items, withIssues);
            if (_this._useList) {
                return ["List<", itemsType, ">"];
            }
            else {
                return [itemsType, "[]"];
            }
        }, function (classType) { return _this.nameForNamedType(classType); }, function (mapType) { return ["Dictionary<string, ", _this.csType(mapType.values, withIssues), ">"]; }, function (enumType) { return _this.nameForNamedType(enumType); }, function (unionType) {
            var nullable = TypeUtils_1.nullableFromUnion(unionType);
            if (nullable !== null)
                return _this.nullableCSType(nullable);
            return _this.nameForNamedType(unionType);
        }, {
            dateTimeType: function (_) { return "DateTimeOffset"; }
        });
    };
    CSharpRenderer.prototype.nullableCSType = function (t, withIssues) {
        if (withIssues === void 0) { withIssues = false; }
        var csType = this.csType(t, withIssues);
        if (isValueType(t)) {
            return [csType, "?"];
        }
        else {
            return csType;
        }
    };
    CSharpRenderer.prototype.superclassForType = function (_t) {
        return undefined;
    };
    CSharpRenderer.prototype.emitType = function (description, accessModifier, declaration, name, superclass, emitter) {
        switch (accessModifier) {
            case AccessModifier.Public:
                declaration = ["public ", declaration];
                break;
            case AccessModifier.Internal:
                declaration = ["internal ", declaration];
                break;
            default:
                break;
        }
        this.emitDescription(description);
        if (superclass === undefined) {
            this.emitLine(declaration, " ", name);
        }
        else {
            this.emitLine(declaration, " ", name, " : ", superclass);
        }
        this.emitBlock(emitter);
    };
    CSharpRenderer.prototype.attributesForProperty = function (_property, _jsonName) {
        return undefined;
    };
    CSharpRenderer.prototype.emitDescriptionBlock = function (lines) {
        var start = "/// <summary>";
        if (this.dense) {
            this.emitLine(start, lines.join("; "), "</summary>");
        }
        else {
            this.emitCommentLines(lines, "/// ", start, "/// </summary>");
        }
    };
    CSharpRenderer.prototype.blankLinesBetweenAttributes = function () {
        return false;
    };
    CSharpRenderer.prototype.emitClassDefinition = function (c, className) {
        var _this = this;
        this.emitType(this.descriptionForType(c), AccessModifier.Public, "partial class", className, this.superclassForType(c), function () {
            if (c.getProperties().isEmpty())
                return;
            var blankLines = _this.blankLinesBetweenAttributes() ? "interposing" : "none";
            var columns = [];
            var isFirstProperty = true;
            var previousDescription = undefined;
            _this.forEachClassProperty(c, blankLines, function (name, jsonName, p) {
                var csType = p.isOptional ? _this.nullableCSType(p.type, true) : _this.csType(p.type, true);
                var attributes = _this.attributesForProperty(p, jsonName);
                var description = _this.descriptionForClassProperty(c, jsonName);
                var property = ["public ", csType, " ", name, " { get; set; }"];
                if (attributes === undefined) {
                    if (
                    // Descriptions should be preceded by an empty line
                    (!isFirstProperty && description !== undefined) ||
                        // If the previous property has a description, leave an empty line
                        previousDescription !== undefined) {
                        _this.ensureBlankLine();
                    }
                    _this.emitDescription(description);
                    _this.emitLine(property);
                }
                else if (_this.dense && attributes.length > 0) {
                    var comment = description === undefined ? "" : " // " + description.join("; ");
                    columns.push([attributes, " ", property, comment]);
                }
                else {
                    _this.emitDescription(description);
                    for (var _i = 0, attributes_1 = attributes; _i < attributes_1.length; _i++) {
                        var attribute = attributes_1[_i];
                        _this.emitLine(attribute);
                    }
                    _this.emitLine(property);
                }
                isFirstProperty = false;
                previousDescription = description;
            });
            if (columns.length > 0) {
                _this.emitTable(columns);
            }
        });
    };
    CSharpRenderer.prototype.emitUnionDefinition = function (u, unionName) {
        var _this = this;
        var nonNulls = TypeUtils_1.removeNullFromUnion(u)[1];
        this.emitType(this.descriptionForType(u), AccessModifier.Public, "partial struct", unionName, this.superclassForType(u), function () {
            _this.forEachUnionMember(u, nonNulls, "none", null, function (fieldName, t) {
                var csType = _this.nullableCSType(t);
                _this.emitLine("public ", csType, " ", fieldName, ";");
            });
        });
    };
    CSharpRenderer.prototype.emitEnumDefinition = function (e, enumName) {
        var caseNames = [];
        this.forEachEnumCase(e, "none", function (name) {
            if (caseNames.length > 0)
                caseNames.push(", ");
            caseNames.push(name);
        });
        this.emitDescription(this.descriptionForType(e));
        this.emitLine("public enum ", enumName, " { ", caseNames, " };");
    };
    CSharpRenderer.prototype.emitExpressionMember = function (declare, define) {
        var _this = this;
        if (this._version === 5) {
            this.emitLine(declare);
            this.emitBlock(function () {
                _this.emitLine("return ", define, ";");
            });
        }
        else {
            this.emitLine(declare, " => ", define, ";");
        }
    };
    CSharpRenderer.prototype.emitTypeSwitch = function (types, condition, withBlock, withReturn, f) {
        var _this = this;
        Support_1.assert(!withReturn || withBlock, "Can only have return with block");
        types.forEach(function (t) {
            _this.emitLine("if (", condition(t), ")");
            if (withBlock) {
                _this.emitBlock(function () {
                    f(t);
                    if (withReturn) {
                        _this.emitLine("return;");
                    }
                });
            }
            else {
                _this.indent(function () { return f(t); });
            }
        });
    };
    CSharpRenderer.prototype.emitUsing = function (ns) {
        this.emitLine("using ", ns, ";");
    };
    CSharpRenderer.prototype.emitUsings = function () {
        for (var _i = 0, _a = ["System", "System.Collections.Generic"]; _i < _a.length; _i++) {
            var ns = _a[_i];
            this.emitUsing(ns);
        }
    };
    CSharpRenderer.prototype.emitRequiredHelpers = function () {
        return;
    };
    CSharpRenderer.prototype.emitTypesAndSupport = function () {
        var _this = this;
        this.forEachObject("leading-and-interposing", function (c, name) { return _this.emitClassDefinition(c, name); });
        this.forEachEnum("leading-and-interposing", function (e, name) { return _this.emitEnumDefinition(e, name); });
        this.forEachUnion("leading-and-interposing", function (u, name) { return _this.emitUnionDefinition(u, name); });
        this.emitRequiredHelpers();
    };
    CSharpRenderer.prototype.emitDefaultLeadingComments = function () {
        return;
    };
    CSharpRenderer.prototype.needNamespace = function () {
        return true;
    };
    CSharpRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        if (this.leadingComments !== undefined) {
            this.emitCommentLines(this.leadingComments);
        }
        else {
            this.emitDefaultLeadingComments();
        }
        this.ensureBlankLine();
        if (this.needNamespace()) {
            this.emitLine("namespace ", this.namespaceName);
            this.emitBlock(function () {
                _this.emitUsings();
                _this.emitTypesAndSupport();
            });
        }
        else {
            this.emitUsings();
            this.emitTypesAndSupport();
        }
    };
    CSharpRenderer.prototype.registerHandlebarsHelpers = function (context) {
        _super.prototype.registerHandlebarsHelpers.call(this, context);
        handlebars.registerHelper("string_escape", Strings_1.utf16StringEscape);
    };
    CSharpRenderer.prototype.makeHandlebarsContextForType = function (t) {
        var ctx = _super.prototype.makeHandlebarsContextForType.call(this, t);
        ctx.csType = this.sourcelikeToString(this.csType(t));
        return ctx;
    };
    CSharpRenderer.prototype.makeHandlebarsContextForUnionMember = function (t, name) {
        var value = _super.prototype.makeHandlebarsContextForUnionMember.call(this, t, name);
        value.nullableCSType = this.sourcelikeToString(this.nullableCSType(t));
        return value;
    };
    return CSharpRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
exports.CSharpRenderer = CSharpRenderer;
var NewtonsoftCSharpRenderer = /** @class */ (function (_super) {
    __extends(NewtonsoftCSharpRenderer, _super);
    function NewtonsoftCSharpRenderer(targetLanguage, graph, leadingComments, namespaceName, version, dense, useList, useDecimal, outputFeatures, _checkRequiredProperties) {
        var _this = _super.call(this, targetLanguage, graph, leadingComments, namespaceName, version, dense, useList, useDecimal) || this;
        _this._checkRequiredProperties = _checkRequiredProperties;
        _this._enumExtensionsNames = immutable_1.Map();
        _this._needHelpers = outputFeatures.helpers;
        _this._needAttributes = outputFeatures.attributes;
        return _this;
    }
    NewtonsoftCSharpRenderer.prototype.forbiddenNamesForGlobalNamespace = function () {
        var forbidden = [
            "Converter",
            "JsonConverter",
            "JsonSerializer",
            "JsonWriter",
            "JsonToken",
            "Serialize",
            "Newtonsoft",
            "MetadataPropertyHandling",
            "DateParseHandling",
            "FromJson",
            "Required"
        ];
        if (this.dense) {
            forbidden.push("J", "R", "N");
        }
        return _super.prototype.forbiddenNamesForGlobalNamespace.call(this).concat(forbidden);
    };
    NewtonsoftCSharpRenderer.prototype.forbiddenForObjectProperties = function (c, className) {
        var result = _super.prototype.forbiddenForObjectProperties.call(this, c, className);
        result.names = result.names.concat(["ToJson", "FromJson", "Required"]);
        return result;
    };
    NewtonsoftCSharpRenderer.prototype.makeNamedTypeDependencyNames = function (t, name) {
        if (!(t instanceof Type_1.EnumType))
            return [];
        var extensionsName = new Naming_1.DependencyName(namingFunction, name.order, function (lookup) { return lookup(name) + "_extensions"; });
        this._enumExtensionsNames = this._enumExtensionsNames.set(name, extensionsName);
        return [extensionsName];
    };
    NewtonsoftCSharpRenderer.prototype.emitUsings = function () {
        // FIXME: We need System.Collections.Generic whenever we have maps or use List.
        if (!this._needAttributes && !this._needHelpers)
            return;
        _super.prototype.emitUsings.call(this);
        this.ensureBlankLine();
        for (var _i = 0, _a = ["System.Globalization", "Newtonsoft.Json", "Newtonsoft.Json.Converters"]; _i < _a.length; _i++) {
            var ns = _a[_i];
            this.emitUsing(ns);
        }
        if (this.dense) {
            this.emitUsing([denseJsonPropertyName, " = Newtonsoft.Json.JsonPropertyAttribute"]);
            this.emitUsing([denseRequiredEnumName, " = Newtonsoft.Json.Required"]);
            this.emitUsing([denseNullValueHandlingEnumName, " = Newtonsoft.Json.NullValueHandling"]);
        }
    };
    NewtonsoftCSharpRenderer.prototype.emitDefaultLeadingComments = function () {
        var _this = this;
        if (!this._needHelpers)
            return;
        this.emitLine("// To parse this JSON data, add NuGet 'Newtonsoft.Json' then do", this.topLevels.size === 1 ? "" : " one of these", ":");
        this.emitLine("//");
        this.emitLine("//    using ", this.namespaceName, ";");
        this.emitLine("//");
        this.forEachTopLevel("none", function (t, topLevelName) {
            var rhs;
            if (t instanceof Type_1.EnumType) {
                rhs = ["JsonConvert.DeserializeObject<", topLevelName, ">(jsonString)"];
            }
            else {
                rhs = [topLevelName, ".FromJson(jsonString)"];
            }
            _this.emitLine("//    var ", Source_1.modifySource(Strings_1.camelCase, topLevelName), " = ", rhs, ";");
        });
    };
    NewtonsoftCSharpRenderer.prototype.attributesForProperty = function (property, jsonName) {
        if (!this._needAttributes)
            return undefined;
        var t = property.type;
        var jsonProperty = this.dense ? denseJsonPropertyName : "JsonProperty";
        var escapedName = Strings_1.utf16StringEscape(jsonName);
        var isNullable = t.isNullable;
        var isOptional = property.isOptional;
        var requiredClass = this.dense ? "R" : "Required";
        var nullValueHandlingClass = this.dense ? "N" : "NullValueHandling";
        var nullValueHandling = isOptional && !isNullable ? [", NullValueHandling = ", nullValueHandlingClass, ".Ignore"] : [];
        var required;
        if (!this._checkRequiredProperties || (isOptional && isNullable)) {
            required = [nullValueHandling];
        }
        else if (isOptional && !isNullable) {
            required = [", Required = ", requiredClass, ".DisallowNull", nullValueHandling];
        }
        else if (!isOptional && isNullable) {
            required = [", Required = ", requiredClass, ".AllowNull"];
        }
        else {
            required = [", Required = ", requiredClass, ".Always", nullValueHandling];
        }
        return [["[", jsonProperty, '("', escapedName, '"', required, ")]"]];
    };
    NewtonsoftCSharpRenderer.prototype.blankLinesBetweenAttributes = function () {
        return this._needAttributes && !this.dense;
    };
    NewtonsoftCSharpRenderer.prototype.emitFromJsonForTopLevel = function (t, name) {
        var _this = this;
        if (t instanceof Type_1.EnumType)
            return;
        var partial;
        var typeKind;
        var definedType = this.namedTypeToNameForTopLevel(t);
        if (definedType !== undefined) {
            partial = "partial ";
            typeKind = definedType instanceof Type_1.ClassType ? "class" : "struct";
        }
        else {
            partial = "";
            typeKind = "class";
        }
        var csType = this.csType(t);
        this.emitType(undefined, AccessModifier.Public, [partial, typeKind], name, this.superclassForType(t), function () {
            // FIXME: Make FromJson a Named
            _this.emitExpressionMember(["public static ", csType, " FromJson(string json)"], ["JsonConvert.DeserializeObject<", csType, ">(json, ", _this.namespaceName, ".Converter.Settings)"]);
        });
    };
    NewtonsoftCSharpRenderer.prototype.emitEnumExtension = function (e, enumName) {
        var _this = this;
        this.emitType(undefined, AccessModifier.None, "static class", Support_1.defined(this._enumExtensionsNames.get(enumName)), this.superclassForType(e), function () {
            _this.emitLine("public static ", enumName, "? ValueForString(string str)");
            _this.emitBlock(function () {
                _this.emitLine("switch (str)");
                _this.emitBlock(function () {
                    _this.forEachEnumCase(e, "none", function (name, jsonName) {
                        _this.emitLine('case "', Strings_1.utf16StringEscape(jsonName), '": return ', enumName, ".", name, ";");
                    });
                    _this.emitLine("default: return null;");
                });
            });
            _this.ensureBlankLine();
            _this.emitLine("public static ", enumName, " ReadJson(JsonReader reader, JsonSerializer serializer)");
            _this.emitBlock(function () {
                _this.emitLine("var str = serializer.Deserialize<string>(reader);");
                _this.emitLine("var maybeValue = ValueForString(str);");
                _this.emitLine("if (maybeValue.HasValue) return maybeValue.Value;");
                _this.emitLine('throw new Exception("Unknown enum case " + str);');
            });
            _this.ensureBlankLine();
            _this.emitLine("public static void WriteJson(this ", enumName, " value, JsonWriter writer, JsonSerializer serializer)");
            _this.emitBlock(function () {
                _this.emitLine("switch (value)");
                _this.emitBlock(function () {
                    _this.forEachEnumCase(e, "none", function (name, jsonName) {
                        _this.emitLine("case ", enumName, ".", name, ': serializer.Serialize(writer, "', Strings_1.utf16StringEscape(jsonName), '"); break;');
                    });
                });
            });
        });
    };
    NewtonsoftCSharpRenderer.prototype.emitUnionJSONPartial = function (u, unionName) {
        var _this = this;
        var tokenCase = function (tokenType) {
            _this.emitLine("case JsonToken.", tokenType, ":");
        };
        var emitNullDeserializer = function () {
            tokenCase("Null");
            _this.indent(function () { return _this.emitLine("return;"); });
        };
        var emitDeserializeType = function (t) {
            _this.emitLine(_this.nameForUnionMember(u, t), " = serializer.Deserialize<", _this.csType(t), ">(reader);");
            _this.emitLine("return;");
        };
        var emitDeserializer = function (tokenType, kind) {
            var t = u.findMember(kind);
            if (t === undefined)
                return;
            tokenCase(tokenType);
            _this.indent(function () { return emitDeserializeType(t); });
        };
        var emitDoubleDeserializer = function () {
            var t = u.findMember("double");
            if (t === undefined)
                return;
            if (u.findMember("integer") === undefined)
                tokenCase("Integer");
            tokenCase("Float");
            _this.indent(function () { return emitDeserializeType(t); });
        };
        var emitStringDeserializer = function () {
            var stringTypes = u.stringTypeMembers;
            if (stringTypes.isEmpty())
                return;
            tokenCase("String");
            tokenCase("Date");
            _this.indent(function () {
                if (stringTypes.size === 1) {
                    emitDeserializeType(Support_1.defined(stringTypes.first()));
                    return;
                }
                _this.emitLine("var str = serializer.Deserialize<string>(reader);");
                _this.forEachUnionMember(u, stringTypes, "none", null, function (fieldName, t) {
                    if (t instanceof Type_1.EnumType) {
                        var extension = Support_1.defined(_this._enumExtensionsNames.get(_this.nameForNamedType(t)));
                        _this.emitLine("var maybeEnum = ", extension, ".ValueForString(str);");
                        _this.emitLine("if (maybeEnum.HasValue)");
                        _this.emitBlock(function () {
                            _this.emitLine(fieldName, " = maybeEnum.Value;");
                            _this.emitLine("return;");
                        });
                    }
                    else if (t.kind === "date-time") {
                        _this.emitLine("DateTimeOffset  dt;");
                        _this.emitLine("if (DateTimeOffset.TryParse(str, out dt))");
                        _this.emitBlock(function () {
                            _this.emitLine(fieldName, " = dt;");
                            _this.emitLine("return;");
                        });
                    }
                    else {
                        return Support_1.panic("Unsupported string enum type " + t.kind);
                    }
                });
                _this.emitLine("break;");
            });
        };
        var _a = TypeUtils_1.removeNullFromUnion(u), hasNull = _a[0], nonNulls = _a[1];
        this.emitType(undefined, AccessModifier.Public, "partial struct", unionName, this.superclassForType(u), function () {
            _this.emitLine("public ", unionName, "(JsonReader reader, JsonSerializer serializer)");
            _this.emitBlock(function () {
                _this.forEachUnionMember(u, nonNulls, "none", null, function (fieldName, _) {
                    _this.emitLine(fieldName, " = null;");
                });
                _this.ensureBlankLine();
                _this.emitLine("switch (reader.TokenType)");
                _this.emitBlock(function () {
                    if (hasNull !== null)
                        emitNullDeserializer();
                    emitDeserializer("Integer", "integer");
                    emitDoubleDeserializer();
                    emitDeserializer("Boolean", "bool");
                    emitDeserializer("StartArray", "array");
                    emitDeserializer("StartObject", "class");
                    emitDeserializer("StartObject", "map");
                    emitStringDeserializer();
                });
                _this.emitLine('throw new Exception("Cannot convert ', unionName, '");');
            });
            _this.ensureBlankLine();
            _this.emitLine("public void WriteJson(JsonWriter writer, JsonSerializer serializer)");
            _this.emitBlock(function () {
                _this.forEachUnionMember(u, nonNulls, "none", null, function (fieldName, _) {
                    _this.emitLine("if (", fieldName, " != null)");
                    _this.emitBlock(function () {
                        _this.emitLine("serializer.Serialize(writer, ", fieldName, ");");
                        _this.emitLine("return;");
                    });
                });
                if (hasNull !== null) {
                    _this.emitLine("writer.WriteNull();");
                }
                else {
                    _this.emitLine('throw new Exception("Union must not be null");');
                }
            });
        });
    };
    NewtonsoftCSharpRenderer.prototype.emitSerializeClass = function () {
        var _this = this;
        // FIXME: Make Serialize a Named
        this.emitType(undefined, AccessModifier.Public, "static class", "Serialize", undefined, function () {
            _this.topLevels.forEach(function (t, _) {
                // FIXME: Make ToJson a Named
                _this.emitExpressionMember(["public static string ToJson(this ", _this.csType(t), " self)"], ["JsonConvert.SerializeObject(self, ", _this.namespaceName, ".Converter.Settings)"]);
            });
        });
    };
    NewtonsoftCSharpRenderer.prototype.emitConverterMembers = function () {
        var _this = this;
        var enumNames = this.enums.map(this.nameForNamedType);
        var unionNames = this.namedUnions.map(this.nameForNamedType);
        var allNames = enumNames.union(unionNames);
        var canConvertExprs = allNames.map(function (n) { return ["t == typeof(", n, ")"]; });
        var nullableCanConvertExprs = allNames.map(function (n) { return ["t == typeof(", n, "?)"]; });
        var canConvertExpr = Support_1.intercalate(" || ", canConvertExprs.union(nullableCanConvertExprs));
        // FIXME: make Iterable<any, Sourcelike> a Sourcelike, too?
        this.emitExpressionMember("public override bool CanConvert(Type t)", canConvertExpr.toArray());
        this.ensureBlankLine();
        this.emitLine("public override object ReadJson(JsonReader reader, Type t, object existingValue, JsonSerializer serializer)");
        this.emitBlock(function () {
            _this.emitTypeSwitch(enumNames, function (t) { return ["t == typeof(", t, ")"]; }, false, false, function (n) {
                var extensionsName = Support_1.defined(_this._enumExtensionsNames.get(n));
                _this.emitLine("return ", extensionsName, ".ReadJson(reader, serializer);");
            });
            _this.emitTypeSwitch(enumNames, function (t) { return ["t == typeof(", t, "?)"]; }, true, false, function (n) {
                _this.emitLine("if (reader.TokenType == JsonToken.Null) return null;");
                var extensionsName = Support_1.defined(_this._enumExtensionsNames.get(n));
                _this.emitLine("return ", extensionsName, ".ReadJson(reader, serializer);");
            });
            // FIXME: call the constructor via reflection?
            _this.emitTypeSwitch(unionNames, function (t) { return ["t == typeof(", t, ") || t == typeof(", t, "?)"]; }, false, false, function (n) {
                _this.emitLine("return new ", n, "(reader, serializer);");
            });
            _this.emitLine('throw new Exception("Unknown type");');
        });
        this.ensureBlankLine();
        this.emitLine("public override void WriteJson(JsonWriter writer, object value, JsonSerializer serializer)");
        this.emitBlock(function () {
            _this.emitLine("var t = value.GetType();");
            _this.emitTypeSwitch(allNames, function (t) { return ["t == typeof(", t, ")"]; }, true, true, function (n) {
                _this.emitLine("((", n, ")value).WriteJson(writer, serializer);");
            });
            _this.emitLine('throw new Exception("Unknown type");');
        });
    };
    NewtonsoftCSharpRenderer.prototype.emitConverterClass = function () {
        var _this = this;
        var jsonConverter = this.haveEnums || this.haveNamedUnions;
        // FIXME: Make Converter a Named
        var converterName = ["Converter"];
        if (jsonConverter)
            converterName = converterName.concat([": JsonConverter"]);
        var staticOrNot = jsonConverter ? "" : "static ";
        this.emitType(undefined, AccessModifier.Internal, [staticOrNot, "class"], converterName, undefined, function () {
            if (jsonConverter) {
                _this.emitConverterMembers();
                _this.ensureBlankLine();
            }
            _this.emitLine("public static readonly JsonSerializerSettings Settings = new JsonSerializerSettings");
            _this.emitBlock(function () {
                _this.emitLine("MetadataPropertyHandling = MetadataPropertyHandling.Ignore,");
                _this.emitLine("DateParseHandling = DateParseHandling.None,");
                if (_this.haveNamedUnions || _this.haveEnums) {
                    _this.emitMultiline("Converters = { \n    new Converter(),\n    new IsoDateTimeConverter { DateTimeStyles = DateTimeStyles.AssumeUniversal }\n},");
                }
                else {
                    _this.emitMultiline("Converters = { \n    new IsoDateTimeConverter { DateTimeStyles = DateTimeStyles.AssumeUniversal }\n},");
                }
            }, true);
        });
    };
    NewtonsoftCSharpRenderer.prototype.emitRequiredHelpers = function () {
        var _this = this;
        if (this._needHelpers) {
            this.forEachTopLevel("leading-and-interposing", function (t, n) { return _this.emitFromJsonForTopLevel(t, n); });
            this.forEachEnum("leading-and-interposing", function (e, n) { return _this.emitEnumExtension(e, n); });
            this.forEachUnion("leading-and-interposing", function (u, n) { return _this.emitUnionJSONPartial(u, n); });
            this.ensureBlankLine();
            this.emitSerializeClass();
        }
        if (this._needHelpers || (this._needAttributes && (this.haveNamedUnions || this.haveEnums))) {
            this.ensureBlankLine();
            this.emitConverterClass();
        }
    };
    NewtonsoftCSharpRenderer.prototype.needNamespace = function () {
        return this._needHelpers || this._needAttributes;
    };
    NewtonsoftCSharpRenderer.prototype.makeHandlebarsContextForType = function (t) {
        var ctx = _super.prototype.makeHandlebarsContextForType.call(this, t);
        if (t.kind === "enum") {
            var name = this.nameForNamedType(t);
            ctx.extensionsName = Support_1.defined(this.names.get(Support_1.defined(this._enumExtensionsNames.get(name))));
        }
        return ctx;
    };
    return NewtonsoftCSharpRenderer;
}(CSharpRenderer));
exports.NewtonsoftCSharpRenderer = NewtonsoftCSharpRenderer;
